/*
#  Copyright 2004 University of Southern California
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  Latest version of this software may be found at:
#      http://www-rcf.usc.edu/~garrick/perl-PBS
#  Please send comments to garrick@usc.edu.
*/


%module "PBS"
%{
#include <pbs_ifl.h>
#include <pbs_error.h>

#ifdef HAVE_RPP_H
#include <rpp.h>
#else
int rpp_open(char *, unsigned int);
int rpp_bind(unsigned int);
int rpp_poll(void);
int rpp_io(void);
int rpp_read(int, void *, int);
int rpp_write(int, void *, int);
int rpp_close(int);
struct  sockaddr_in* rpp_getaddr(int);
int rpp_flush(int);
void rpp_shutdown(void);
void rpp_terminate(void);
int rpp_rcommit(int, int);
int rpp_wcommit(int, int);
int rpp_skip(int, int);
int rpp_eom(int);
int rpp_getc(int);
int rpp_putc(int, int);
char* netaddr(struct sockaddr_in *);
unsigned long crc(unsigned char *, unsigned long);
extern int rpp_fd;
extern int rpp_dbprt;
extern int rpp_retry;
#endif

#ifdef HAVE_RM_H
#include <rm.h>
#else
int openrm(char *, unsigned int);
int closerm(int);
int downrm(int);
int configrm(int, char *);
int addreq(int, char *);
int allreq(char *);
char* getreq(int);
int flushreq(void);
int activereq(void);
void fullresp(int);
#endif

int cnt2server( char *server );
%}

#ifdef HAVE_RM_H
#include <rm.h>
#else
int     openrm          (char *, unsigned int);
int     closerm         (int);
int     downrm          (int);
int     configrm        (int, char *);
int     addreq          (int, char *);
int     allreq          (char *);
char*   getreq          (int);
int     flushreq        (void);
int     activereq       (void);
void    fullresp        (int);
#endif



%typemap(in)  struct attrl *attrib {
    SV *tempsv;
    AV *tempav;
    HV *temphv;
    HE *hashentry;
    I32 len;
    SV **tv;
    char *keystring;
    SV *th, *valsv;
    int i;
    char *tmpstr;
    char *dotptr;
    struct attrl *attrhead;
    struct attrl *attrptr;
    size_t dotloc;

    if (!SvOK((SV*) $input)) {
        attrhead = NULL;

    } else if ( SvOK((SV*) $input) && (SvTYPE($input) == SVt_PV) ) {
        /* it's one string */
        attrhead=calloc(1, sizeof(struct attrl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        tmpstr=SvPV_nolen($input);
        attrhead->name=strdup(tmpstr);
        dotptr=strchr(attrhead->name, '.');
        if (dotptr!=NULL) {
           attrhead->resource=dotptr+1;
           *dotptr=0;
        }

    } else if ( SvOK((SV*) $input) && (SvTYPE(SvRV($input)) == SVt_PVAV) ) {
        /* it's an array of strings */

        attrhead=calloc(1, sizeof(struct attrl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        attrptr=attrhead;
    
        /* build up the linked list */
        tempav=(AV*)SvRV($input);
        len=av_len(tempav);
    
        for (i = 0; i <= len; i++) {
            tv = av_fetch(tempav, i, 0);
    
            tmpstr=SvPV_nolen(*tv);
            attrptr->name=strdup(tmpstr);
            dotptr=strchr(attrptr->name, '.');
            if (dotptr!=NULL) {
               attrptr->resource=dotptr+1;
               *dotptr=0;
            }
            if (i<len) {
               attrptr->next=calloc(1, sizeof(struct attrl));
               if (attrptr->next==NULL) {
                   fprintf(stderr, "perl-PBS: out of memory\n");
                   return;
               }
               attrptr=attrptr->next;
            }
        }
        
    } else if ( SvOK((SV*) $input) && (SvTYPE(SvRV($input)) == SVt_PVHV) ) {
        /* it's a hash of strings */
        attrhead=calloc(1, sizeof(struct attrl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        attrptr=attrhead;
    
        /* build up the linked list */
        temphv=(HV*)SvRV($input);
        hv_iterinit(temphv);

        hashentry=hv_iternext(temphv);
        keystring=hv_iterkey(hashentry, &len);
        valsv=hv_iterval(temphv, hashentry);

        attrptr->name=strdup(keystring);
        dotptr=strchr(attrptr->name, '.');
        if (dotptr!=NULL) {
           attrptr->resource=dotptr+1;
           *dotptr=0;
        }

        tmpstr=SvPV_nolen(valsv);
        attrptr->value=strdup(tmpstr);

        while(hashentry=hv_iternext(temphv)) {
            attrptr->next=calloc(1, sizeof(struct attrl));
            if (attrptr->next==NULL) {
                fprintf(stderr, "perl-PBS: out of memory\n");
                return;
            }
            attrptr=attrptr->next;

            keystring=hv_iterkey(hashentry, &len);
            valsv=hv_iterval(temphv, hashentry);

            attrptr->name=strdup(keystring);
            dotptr=strchr(attrptr->name, '.');
            if (dotptr!=NULL) {
               attrptr->resource=dotptr+1;
               *dotptr=0;
            }

            tmpstr=SvPV_nolen(valsv);
            attrptr->value=strdup(tmpstr);
        }
        
    } else {
        croak("Unexpected input");
    }
    $1 = attrhead;
    
}

%typemap(in)  struct attropl *attrib {
    SV *tempsv;
    AV *tempav;
    HV *temphv;
    HE *hashentry;
    I32 len;
    SV **tv;
    char *keystring;
    SV *th, *valsv;
    int i;
    char *tmpstr;
    char *dotptr;
    struct attropl *attrhead;                                                              
    struct attropl *attrptr;                                                               
    size_t dotloc;

    if (!SvOK((SV*) $input)) {
        attrhead = NULL;

    } else if ( SvOK((SV*) $input) && (SvTYPE($input) == SVt_PV) ) {
        /* it's one string */
        attrhead=calloc(1, sizeof(struct attropl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        tmpstr=SvPV_nolen($input);
        attrhead->name=strdup(tmpstr);
        dotptr=strchr(attrptr->name, '.');
        if (dotptr!=NULL) {
           attrptr->resource=dotptr+1;
           *dotptr=0;
        }

    } else if ( SvOK((SV*) $input) && (SvTYPE(SvRV($input)) == SVt_PVAV) ) {
        /* it's an array of strings */

        attrhead=calloc(1, sizeof(struct attropl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        attrptr=attrhead;

        /* build up the linked list */
        tempav=(AV*)SvRV($input);
        len=av_len(tempav);

        for (i = 0; i <= len; i++) {
            tv = av_fetch(tempav, i, 0);

            tmpstr=SvPV_nolen(*tv);
            attrptr->name=strdup(tmpstr);
            dotptr=strchr(attrptr->name, '.');
            if (dotptr!=NULL) {
               attrptr->resource=dotptr+1;
               *dotptr=0;
            }
            if (i<len) {
               attrptr->next=calloc(1, sizeof(struct attropl));
               if (attrptr->next==NULL) {
                   fprintf(stderr, "perl-PBS: out of memory\n");
                   return;
               }
               attrptr=attrptr->next;
            }
        }

    } else if ( SvOK((SV*) $input) && (SvTYPE(SvRV($input)) == SVt_PVHV) ) {
        /* it's a hash of strings */
        attrhead=calloc(1, sizeof(struct attropl));
        if (attrhead==NULL) {
            fprintf(stderr, "perl-PBS: out of memory\n");
            return;
        }
        attrptr=attrhead;

        /* build up the linked list */
        temphv=(HV*)SvRV($input);
        hv_iterinit(temphv);

        hashentry=hv_iternext(temphv);
        keystring=hv_iterkey(hashentry, &len);
        valsv=hv_iterval(temphv, hashentry);

        attrptr->name=strdup(keystring);
        dotptr=strchr(attrptr->name, '.');
        if (dotptr!=NULL) {
           attrptr->resource=dotptr+1;
           *dotptr=0;
        }

        tmpstr=SvPV_nolen(valsv);
        attrptr->value=strdup(tmpstr);

        while(hashentry=hv_iternext(temphv)) {
            attrptr->next=calloc(1, sizeof(struct attropl));
            if (attrptr->next==NULL) {
                fprintf(stderr, "perl-PBS: out of memory\n");
                return;
            }
            attrptr=attrptr->next;

            keystring=hv_iterkey(hashentry, &len);
            valsv=hv_iterval(temphv, hashentry);

            attrptr->name=strdup(keystring);

            dotptr=strchr(attrptr->name, '.');
            if (dotptr!=NULL) {
               attrptr->resource=dotptr+1;
               *dotptr=0;
            }
            tmpstr=SvPV_nolen(valsv);
            attrptr->value=strdup(tmpstr);
        }

    } else {
        croak("Unexpected input");
    }
#ifdef DEBUGPBS
    attrptr=attrhead;
    while(attrptr!=NULL) {
       printf ("name=%s, resource=%s, value=%s\n", 
          attrptr->name!=NULL ? attrptr->name : "NULL",
          attrptr->resource!=NULL ? attrptr->resource : "NULL",
          attrptr->value!=NULL ? attrptr->value : "NULL"
       );
      attrptr=attrptr->next;
    }
#endif
    $1 = attrhead;

}


%typemap(freearg) struct attrl * {
    struct attrl *freeattrptr;
    struct attrl *freeattrnext;

    freeattrptr=$1;

    while (freeattrptr!=NULL) {
        freeattrnext=freeattrptr->next;

        /* resource isn't free'd because it was never malloc'd */
        if (freeattrptr->name != NULL) 
            free(freeattrptr->name);
        if (freeattrptr->value != NULL) 
            free(freeattrptr->value);

        free(freeattrptr);
        freeattrptr=freeattrnext;
    }
}

%typemap(freearg) struct attropl * {
    struct attropl *freeattrptr;
    struct attropl *freeattrnext;

    freeattrptr=$1;

    while (freeattrptr!=NULL) {
        freeattrnext=freeattrptr->next;

        /* resource isn't free'd because it was never malloc'd */
        if (freeattrptr->name != NULL) 
            free(freeattrptr->name);
        if (freeattrptr->value != NULL) 
            free(freeattrptr->value);

        free(freeattrptr);
        freeattrptr=freeattrnext;
    }
}

%typemap(perl5,out)  struct batch_status * {
    AV *newresult;
    int b=0, a=0;
    struct batch_status *batchptr, *batchhead;

    HV*  newhash;
    struct attrl *attrptr, *attrfree;
    time_t epoch;
    char *tmpstr;

    AV *attribarray;
    HV *attribhash;

    newresult=newAV();
    batchptr=$1;
    batchhead=$1;
    while (batchptr!=NULL) {

        newhash=newHV();
        hv_store(newhash, "name", 4, newSVpv( batchptr->name, strlen(batchptr->name)), 0);

        if (batchptr->text!=NULL)
           hv_store(newhash, "text", 4, newSVpv( batchptr->text, strlen(batchptr->text)), 0);

        if (batchptr->attribs!=NULL) {
           attribarray=newAV();

           attrptr=batchptr->attribs;

           while (attrptr!=NULL) {
              attribhash=newHV();

              if (attrptr->name!=NULL) {
                  if (attrptr->resource!=NULL) {
                      tmpstr=malloc(sizeof(char)* MAXNAMLEN+MAXNAMLEN+2);
                      if(tmpstr==NULL) {
                         fprintf(stderr, "perl-PBS: out of memory\n");
                         pbs_statfree(batchhead);
                         return;
                      }
                      sprintf(tmpstr, "%s.%s", attrptr->name, attrptr->resource);
                      hv_store(attribhash, "name", 4, 
                           newSVpv(tmpstr, strlen(tmpstr)),0);
                      free(tmpstr);
                  } else {
                      hv_store(attribhash, "name", 4, newSVpv(attrptr->name, strlen(attrptr->name)),0);
                  }

              }

              if (attrptr->value!=NULL) {
                    if (strcmp(attrptr->name, ATTR_a)==0 ||
                        strcmp(attrptr->name, ATTR_ctime)==0 ||
                        strcmp(attrptr->name, ATTR_etime)==0 ||
                        strcmp(attrptr->name, ATTR_mtime)==0 ||
                        strcmp(attrptr->name, ATTR_qtime)==0 ) {

                        epoch = (time_t) atoi(attrptr->value);
                        tmpstr=ctime(&epoch); /* FIXME non-reentrant ctime with threaded perl */
                        strtok(tmpstr, "\n");
                        hv_store(attribhash, "value", 5, newSVpv(tmpstr, strlen(tmpstr)),0);

                    } else {
                        hv_store(attribhash, "value", 5, 
                                 newSVpv(attrptr->value, strlen(attrptr->value)),0);
                    }
              }

              av_push(attribarray, newRV_noinc((SV*)attribhash));
              attrptr=attrptr->next;
           }

           hv_store(newhash, "attribs", 7, newRV_noinc((SV*)attribarray), 0);
        }
        
        av_push(newresult, newRV_noinc((SV*)newhash));
        batchptr=batchptr->next;
        
    }
    if (batchhead!=NULL)
        pbs_statfree(batchhead);

    $result = newRV_noinc((SV*)newresult);
    sv_2mortal($result);  
    argvi++;

}


%ignore pbs_statdest;
%include <pbs_ifl.h>

#ifdef BROKEN_PBSE_TO_TXT
char *pbse_to_txt( int err);
%ignore pbse_to_txt;
#endif

%ignore pbs_errno;
%include <pbs_error.h>

#ifdef HAVE_RPP_H
%include <rpp.h>
#else
int rpp_open (char *, unsigned int);
int rpp_bind (unsigned int);
int rpp_poll (void);
int rpp_io (void);
int rpp_read (int, void *, int);
int rpp_write (int, void *, int);
int rpp_close (int);
struct  sockaddr_in* rpp_getaddr (int);
int rpp_flush (int);
void rpp_shutdown (void);
void rpp_terminate (void);
int rpp_rcommit (int, int);
int rpp_wcommit (int, int);
int rpp_skip (int, int);
int rpp_eom (int);
int rpp_getc (int);
int rpp_putc (int, int);
char* netaddr (struct sockaddr_in *);
unsigned long crc (unsigned char *, unsigned long);
extern  int     rpp_fd;
extern  int     rpp_dbprt;
extern  int     rpp_retry;
#endif

int cnt2server( char *server );


